Dir["../site/*"].each{|filename|
  if !File.exist?(uploaded_filename=filename.sub("site","site_uploaded")) || File.read(filename) != File.read(uploaded_filename)
    `putweb golfscript.com public_html/iogii/ #{filename} && mv #{filename} #{uploaded_filename}`
  else
    `rm #{filename}`
  end
}
raise if !Dir["../site/*"].empty?
