export LANG=iso-8859-1

iogii test/lang_bytes_prog.iog < test/lang_bytes_input.txt > output.txt
diff output.txt test/lang_bytes_expected.txt

ghc -O engine.hs
iogii test/lang_bytes_prog.iog -e > t.ism && engine t.ism < test/lang_bytes_input.txt > output.txt
diff output.txt test/lang_bytes_expected.txt
